@@ -106,7 +106,10 @@ class Algorithms:
         orgh, orgw = frame.shape[:2]
         dsth, dstw = int(target_height), int(target_width)
         if (dstw, dsth) != (orgw, orgh):
+            dt = frame.dtype
+            frame = frame.astype(np.float32) if dt == np.float16 else frame
             frame = cv2.resize(frame, (dstw, dsth), cv2.INTER_LANCZOS4)
+            frame = frame.astype(dt)
             dsth, dstw = frame.shape[:2]
             self._vprint(f"{_elapsed(t0)} - resizing [Lanczos] from {orgw} x {orgh} to {dstw} x {dsth}")
         return frame
@@ -119,7 +122,7 @@ class Algorithms:
         that the per-channel levels are in the same Bayer order as the frame itself.
         """
         t0 = time.time()
-        levels = np.array(blacklevels)
+        levels = np.asarray(blacklevels, dtype=frame.dtype)
         if frame.ndim == 3:  # already demosaiced
             assert levels.size == 3, levels
             frame = np.maximum(frame, levels)
@@ -159,9 +162,10 @@ class Algorithms:
             self._vprint(f"{_elapsed(t0)} - estimating white level: {percentile:5.2f}th percentile = {whitelevel:.2f}")
         assert whitelevel > blacklevel, f"{whitelevel} is not greater than {blacklevel}"
         frame = np.clip(frame, blacklevel, whitelevel)
+        frame = frame.astype(np.float32)
         t0 = time.time()
         frame -= blacklevel
-        frame = frame.astype(np.float32) / (whitelevel - blacklevel)
+        frame = frame / (whitelevel - blacklevel)
         ranges = f"{minmax} => [{blacklevel:.2f}, {whitelevel:.2f}] => {self._minmax(frame)}"
         self._vprint(f"{_elapsed(t0)} - linearizing: range = {ranges}")
         return frame
@@ -188,9 +192,11 @@ class Algorithms:
                             "GBRG": cv2.COLOR_BAYER_GR2RGB_EA,
                             "BGGR": cv2.COLOR_BAYER_RG2RGB_EA,
                             "GRBG": cv2.COLOR_BAYER_GB2RGB_EA}
+            dt = frame.dtype
             frame = np.rint(frame * 65535).astype(np.uint16)
             frame = cv2.cvtColor(frame, bayer_to_cv2[bayer_pattern.upper()])
             frame = frame / 65535.0
+            frame = frame.astype(dt)
             method = "EAD"
         else:
             channels = self.bayer_split(frame)
@@ -223,13 +229,13 @@ class Algorithms:
             lsch, lscw = lscmap.shape[:2]
             need_resize = lscmap.shape[:2] != frame.shape[:2]
             if need_resize:
+                dt = lscmap.dtype
                 xgrid = np.linspace(0, lscw - 1, imgw)
                 ygrid = np.linspace(0, lsch - 1, imgh)
                 mgrid = np.dstack(np.meshgrid(xgrid, ygrid, indexing="xy"))
-                numthreads = cv2.getNumThreads()
-                cv2.setNumThreads(0)  # work-around for OpenCV 4.1.2 deadlock
+                lscmap = lscmap.astype(np.float32) if dt == np.float16 else lscmap
                 lscmap = cv2.remap(lscmap, mgrid.astype(np.float32), None, cv2.INTER_LINEAR)
-                cv2.setNumThreads(numthreads)
+                lscmap = lscmap.astype(dt)
             if lscmap.ndim < frame.ndim:
                 lscmap = np.atleast_3d(lscmap)  # {RGB, monochrome} => RGB
             frame = frame * lscmap
@@ -250,7 +256,7 @@ class Algorithms:
         """
         if gains is not None:
             t0 = time.time()
-            wb = np.asarray(gains)
+            wb = np.asarray(gains, dtype=frame.dtype)
             wb = np.insert(wb, 1, 1.0) if len(wb) == 2 else wb
             if frame.ndim == 3:  # RGB mode
                 mode = "RGB"
@@ -283,6 +289,7 @@ class Algorithms:
             if clip:
                 frame = self.clip(frame, 0, 1)
             t0 = time.time()
+            matrix = matrix.astype(frame.dtype)
             frame = frame @ matrix.T
             with np.printoptions(formatter={'float': lambda x: f"{x:.2f}"}):
                 sums = f"with column sums {np.sum(matrix, axis=0).T}"
@@ -300,7 +307,10 @@ class Algorithms:
         if mode == "ACES":
             assert p >= 1, f"compression curve power must be >= 1.0; was {p:.3f}"
             t0 = time.time()
+            dt = frame.dtype
+            frame = frame.astype(np.float32) if dt == np.float16 else frame
             frame = _aces_gamut(frame, power=p)
+            frame = frame.astype(dt)
             self._vprint(f"{_elapsed(t0)} - gamut mapping [{mode}, p={p:.2f}]: range = {self._minmax(frame)}")
         return frame
 
@@ -313,9 +323,13 @@ class Algorithms:
         if mode == "Reinhard":
             frame = self.clip(frame, 0, np.inf)
             t0 = time.time()
-            frame = frame.astype(np.float32)  # can't handle float64
+            dt = frame.dtype
+            frame = frame.astype(np.float32)  # can't handle any other dtypes
             algo = cv2.createTonemapReinhard(gamma=1.0, intensity=0.0, light_adapt=0.0, color_adapt=0.0)
-            frame = algo.process(frame)
+            cv2.setLogLevel(2)  # 2 = LOG_LEVEL_ERROR
+            frame = algo.process(frame)  # causes a spurious internal warning in OpenCV 4.5
+            cv2.setLogLevel(3)  # 3 = LOG_LEVEL_WARNING
+            frame = frame.astype(dt)
             self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {self._minmax(frame)}")
         return frame
 
@@ -353,10 +367,13 @@ class Algorithms:
         """
         if booster is not None:
             t0 = time.time()
+            dt = frame.dtype
+            frame = frame.astype(np.float32) if dt == np.float16 else frame
             hsl = _transform_srgb_to_hsl(frame)  # HSL is in planar form
             hsl[1] = booster(hsl[1])  # saturation boost
             hsl[1] = np.clip(hsl[1], 0, 1)  # may be -eps or 1+eps
             frame = _transform_hsl_to_srgb(hsl)  # RGB back in interleaved form
+            frame = frame.astype(dt)
             self._vprint(f"{_elapsed(t0)} - applying HSL saturation boost: range = {self._minmax(frame)}")
         return frame
 
@@ -375,6 +392,7 @@ class Algorithms:
         assert mode in ["sRGB", "rec709", "LUT"] or not mode, f"Unrecognized mode '{repr(mode)}'"
         if mode in ["sRGB", "rec709", "LUT"]:
             t0 = time.time()
+            dt = frame.dtype
             realmode = mode
             input_range = self._minmax(frame)
             frame = np.clip(frame, 0, 1)  # can't handle values outside of [0, 1]
@@ -399,10 +417,11 @@ class Algorithms:
                     lut = lut * maxval
             if mode == "LUT":
                 assert lut is not None
+                lut = lut.astype(dt)
                 maxval = len(lut) - 1
                 frame = self._quantize(frame, maxval)  # [0, 1] ==> [0, maxval]
                 frame = lut[frame]                     # [0, maxval] ==> [0, maxval]
-                frame = frame / maxval                 # [0, maxval] ==> [0, 1]
+                frame = frame / float(maxval)          # [0, maxval] ==> [0, 1]
             self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: range = {input_range} => {self._minmax(frame)}")
         return frame
 
