@@ -27,7 +27,7 @@ class RegressionTest(unittest.TestCase):
         alg = rawpipe.Algorithms(verbose=True)
         raw = alg.subtract(raw, [10, 20, 30, 40])
         raw = alg.subtract(raw, [246, 236, 226, 216])  # total = 256
-        raw = alg.linearize(raw, 0, whitelevel)
+        raw = alg.linearize(raw, 0, whitelevel).astype(np.float64)
         raw = alg.bayer_combine(*alg.bayer_split(raw))  # no-op
         raw = alg.demosaic(raw, bayer_pattern)
         raw = alg.subtract(raw, [0, 0, 0])  # no-op
@@ -51,6 +51,42 @@ class RegressionTest(unittest.TestCase):
         self.assertEqual(res.dtype, np.float32)
         np.testing.assert_allclose(org_b * wb[-1], res_b)
 
+    def test_preserve_dtype(self):
+        raw = np.random.random((28, 12))
+        rgb = np.random.random((28, 12, 3))
+        for dt in [np.float16, np.float32, np.float64]:
+            raw_src = raw.astype(dt)
+            rgb_src = rgb.astype(dt)
+            res01 = rawpipe.clip(raw_src, 0.1, 0.9)
+            res02 = rawpipe.bayer_combine(*rawpipe.bayer_split(raw_src))
+            res03 = rawpipe.downsample(raw_src, 1)
+            res04 = rawpipe.downsample(rgb_src, 1)
+            res05 = rawpipe.resize(rgb_src, 20, 14)
+            res06 = rawpipe.subtract(raw_src, blacklevels=[0.1, 0.2, 0.3, 0.4])
+            res07 = rawpipe.subtract(rgb_src, blacklevels=[0.1, 0.2, 0.3])
+            res08 = rawpipe.linearize(raw_src, 0.1, 0.9)
+            res09 = rawpipe.linearize(rgb_src, 0.1, num_clipped=2)
+            res10 = rawpipe.wb(raw_src, [1.2, 2.9], "RGGB")
+            res11 = rawpipe.wb(rgb_src, [1.2, 1.3, 2.9], "GBRG")
+            res12 = rawpipe.demosaic(raw_src, "RGGB")
+            res13 = rawpipe.demosaic(raw_src, "RGGB", downsample=True)
+            res14 = rawpipe.lsc(raw_src, raw_src[::3, ::2] * 0.5)
+            res15 = rawpipe.lsc(rgb_src, rgb_src[::2, ::3] * 0.5)
+            res16 = rawpipe.ccm(rgb_src, np.arange(9).reshape(3, 3))
+            res17 = rawpipe.gamut(rgb_src)
+            res18 = rawpipe.tonemap(rgb_src)
+            res19 = rawpipe.saturate(rgb_src, lambda v: v ** 0.7)
+            res20 = rawpipe.gamma(raw_src)
+            res21 = rawpipe.gamma(rgb_src)
+            results = [res01, res02, res03, res04, res05, res06,
+                       res07, res10, res11, res12, res13, res14,
+                       res15, res16, res17, res18, res19, res20,
+                       res21]
+            for res in results:
+                self.assertEqual(res.dtype, dt)
+            self.assertEqual(res08.dtype, np.float32)  # linearize()
+            self.assertEqual(res09.dtype, np.float32)  # linearize()
+
     def test_errors(self):
         raw = np.fromfile(os.path.join(thisdir, "input.raw"), dtype=np.uint16)
         res = rawpipe.verbose.gamma(raw, None)  # should be no-op
